
from django.test import TestCase
from django.core.urlresolvers import reverse
from django.http import HttpResponse
from django.conf.urls import patterns, include, url

from viewrouter.routers import Router, route
from viewrouter.views import ActionView

class TestView(ActionView):
    urls = [
        ('^retrieve/%s/pass/$', 'retrieve', 'retrieve'),
    ]
    urls = []

    def index(self, request):
        return HttpResponse('index')

    def create(self, request):
        return HttpResponse('create')

    def retrieve(self, request, pk):
        return HttpResponse('hello %d' % int(pk))

    def update(self, request, pk):
        return HttpResponse('update %d' % int(pk))

    def delete(self, request, pk):
        return HttpResponse('delete %d' % int(pk))

test_router = Router(TestView)
urlpatterns = patterns('', url(r'', include(test_router.urls)))

class TestRouting(TestCase):
    urls = urlpatterns

    def test_view(self):
        url = reverse('testview:retrieve', kwargs={'pk': '2'})
        assert url == '/retrieve/2/', url
        url = reverse('testview:update', kwargs={'pk': '2'})
        assert url == '/update/2/', url
        url = reverse('testview:delete', kwargs={'pk': '2'})
        assert url == '/delete/2/', url
        url = reverse('testview:index')
        assert url == '/', url
        url = reverse('testview:create')
        assert url == '/create/', url

class TestRoutingOverride(TestCase):
    class TestView(ActionView):
        urls = [
            ('^retrieve/(?P<pk>\d+)/pass/$', 'retrieve', 'retrieve', []),
            ('^update/(?P<pk>\d+)/$', 'update', 'update', ['post']),
        ]

        def retrieve(self, request, pk):
            return HttpResponse('hello %d' % int(pk))

        def update(self, request, pk):
            return HttpResponse('hello %d' % int(pk))

    test_router = Router(TestView)
    urlpatterns = patterns('', url(r'', include(test_router.urls)))
    urls = urlpatterns

    def test_view(self):
        url = reverse('testview:retrieve', kwargs={'pk': '2'})
        assert url == '/retrieve/2/pass/', url

    def test_view_post_only(self):
        url = reverse('testview:update', kwargs={'pk': '2'})
        assert url == '/update/2/', url

        resp = self.client.post(url, data={'name': 'test'})
        assert resp.status_code == 200, resp.status_code

        resp = self.client.get(url)
        assert resp.status_code == 405, resp.status_code

class TestRoutingDecorator(TestCase):
    class TestView(ActionView):

        @route(r'^retrieve/(?P<id>\d+)/pass/$')
        def retrieve(self, request, id):
            return HttpResponse('hello %d' % int(id))

        @route(r'^update/(?P<id>\d+)/pass/$', http_methods=['post'])
        def update(self, request, id):
            return HttpResponse('update %d' % int(id))

        @route(http_methods=['get'])
        def index(self, request):
            return HttpResponse('index')

        @route(http_methods=['post'])
        def delete(self, request, pk):
            return HttpResponse('delete %s' % pk)

        @route(http_methods=['post'])
        def set_password(self, request):
            return HttpResponse('set password')

        @route(r'^filter-group/(?P<pk>\d+)/group/(?P<group>\d+)/$')
        def filter_group(self, pk, group=None):
            return HttpResponse('filter_group %s %s' % (pk, group))

    test_router = Router(TestView)
    urlpatterns = patterns('', url(r'', include(test_router.urls)))
    urls = urlpatterns

    def test_view(self):
        url = reverse('testview:retrieve', kwargs={'id': '2'})
        assert url == '/retrieve/2/pass/', url

    def test_view_post_only(self):
        url = reverse('testview:update', kwargs={'id': '2'})
        assert url == '/update/2/pass/', url

        resp = self.client.post(url, data={'name': 'test'})
        assert resp.status_code == 200, resp.status_code

        resp = self.client.get(url)
        assert resp.status_code == 405, resp.status_code

    def test_view_decorate_method_only(self):
        url = reverse('testview:index')
        assert url == '/', url

        resp = self.client.post(url, data={'name': 'test'})
        assert resp.status_code == 405, resp.status_code

        resp = self.client.get(url)
        assert resp.status_code == 200, resp.status_code

    def test_view_decorate_method_only_has_pk(self):
        url = reverse('testview:delete', kwargs={'pk': 2})
        assert url == '/delete/2/', url

        resp = self.client.post(url, data={'name': 'test'})
        assert resp.status_code == 200, resp.status_code

        resp = self.client.get(url)
        assert resp.status_code == 405, resp.status_code

    def test_view_decorate_custom_action(self):
        url = reverse('testview:set_password')
        assert url == '/set_password/', url

    def test_view_decorate_custom_action_with_pattern(self):
        url = reverse('testview:filter_group', kwargs={'pk': 2, 'group': 4})
        assert url == '/filter-group/2/group/4/', url



"""
Copyright (c) 2014 Mohd. Kamal Bin Mustafa

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from django.http import HttpResponse, HttpResponseNotFound
from django.views.generic.base import View
from django.shortcuts import render

class ActionView(View):
    route_action = 'index'
    urls = []
    template_prefix = None
    model = None

    def dispatch(self, request, *args, **kwargs):
        if request.method.lower() in self.http_method_names:
            handler = getattr(self, self.route_action, self.not_found)
        else:
            handler = self.http_method_not_allowed

        response = handler(request, *args, **kwargs)
        if getattr(response, "is_rendered", None):
            # assume a TemplateResponse
            if not response.template_name:
                response.template_name = self.get_template()
        return response

    def not_found(self, *args, **kwargs):
        return HttpResponseNotFound()

    def get_template(self):
        app_name = self.request.resolver_match.app_name
        class_name = self.__class__.__name__.lower()
        template_name = f"{class_name}_{self.route_action}.html"
        prefix = None
        templates = []

        if self.model:
            prefix = self.model._meta.app_label + "/"
        if self.template_prefix:
            prefix = self.template_prefix

        if prefix:
            templates.append(f"{prefix}{template_name}")
            templates.append(f"{prefix}actionview_{self.route_action}.html")
        templates.append(template_name)
        templates.append(f"viewrouter/actionview_{self.route_action}.html")
        return templates

    def get_object(self, *args, **kwargs):
        if not self.model:
            raise Exception("model not defined")

        return self.model.objects.get(*args, **kwargs)

    def index(self, *args, **kwargs):
        return render(self.request, self.get_template())

    @classmethod
    def get_urls(cls):
        from .routers import Router
        return Router(cls).urls



"""
Copyright (c) 2014 Mohd. Kamal Bin Mustafa

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""

from django.urls import include, path

class Router(object):
    action_allowed = ['index', 'create', 'retrieve', 'update', 'delete']
    action_has_pk = ['retrieve', 'update', 'delete']
    
    def __init__(self, view, urls=None):
        self.view = view
        self.view_name = view.__name__.lower()
        self.urlpatterns = urls

    def build_urlpatterns(self, pattern, action, urlname, http_methods):
        kwargs = {
            'route_action': action,
        }
        if len(http_methods) > 0:
            kwargs['http_method_names'] = http_methods
        urlpatterns = [
                path(pattern, self.view.as_view(**kwargs), name=urlname)
            ]
        return urlpatterns
    
    def build_default_pattern(self, action):
        if action in self.action_has_pk:
            pattern = f'{action}/<int:pk>/'
        else:
            pattern = f'{action}/'
        return pattern
        
    @property
    def urls(self):
        urlpatterns = []
        overidden_actions = []
        for _url in self.view.urls:
            pattern, action, urlname, http_methods = _url
            overidden_actions.append(action)
            urlpatterns += self.build_urlpatterns(
                    pattern,
                    action,
                    urlname, 
                    http_methods
                )

        for action in dir(self.view):
            action_callable = getattr(self.view, action, None)
            if action not in self.action_allowed:
                if not hasattr(action_callable, '_route'):
                    continue
            if action in overidden_actions:
                continue

            _urlname, _http_methods = None, []
            if (action_callable is not None and
                    hasattr(action_callable, '_route')):
                pattern, _urlname, _http_methods = action_callable._route
            else:
                pattern = self.build_default_pattern(action)

            # pattern still None, a case when user using @route decorator
            # but does not specify pattern
            if pattern is None:
                pattern = self.build_default_pattern(action)

            as_view_kwargs = {
                'route_action': action,
            }
            urlpatterns += self.build_urlpatterns(
                            pattern,
                            action,
                            _urlname or action,
                            _http_methods
                        )

            # allow index to be accessed as /
            if action == 'index':
                urlpatterns += self.build_urlpatterns(
                            "",
                            action,
                            _urlname or action,
                            _http_methods
                        )
        return (urlpatterns, self.view_name)

def route(pattern=None, name=None, http_methods=None):
    def decorator(fn):
        fn._route = (pattern, name or fn.__name__, http_methods or [])
        return fn
    return decorator
